import pyterrier as pt
from transformers import PreTrainedModel, PretrainedConfig, PreTrainedTokenizer, AutoModelForSequenceClassification, AutoTokenizer, AutoConfig
import torch
import torch.nn.functional as F

class CatConfig(PretrainedConfig):
    model_type = 'Cat'

class Cat(PreTrainedModel):
    """Wrapper for Cat Model
    
    Parameters
    ----------
    model : PreTrainedModel
        the underlying HF model
    config : AutoConfig
        the configuration for the model
    """
    model_type = 'Cat'
    architecture_class = AutoModelForSequenceClassification
    config_class = CatConfig
    transformer_class = None
    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizer,
        config: AutoConfig,
    ):
        super().__init__(config)
        self.model = model
        self.tokenizer = tokenizer

        from .pyterrier.cat import CatTransformer
        self.transformer_class = CatTransformer
    
    def prepare_outputs(self, logits, labels=None):
        """Prepare outputs"""
        return F.log_softmax(logits.reshape(-1, self.config.group_size, 2), dim=-1)[:, :, 1], labels.view(-1, self.config.group_size) if labels is not None else None

    def forward(self, loss, sequences, labels=None):
        """Compute the loss given (pairs, labels)"""
        sequences = {k: v.to(self.model.device) for k, v in sequences.items()}
        labels = labels.to(self.model.device) if labels is not None else None
        logits = self.model(**sequences).logits
        pred, labels = self.prepare_outputs(logits, labels)
        loss_value = loss(pred) if labels is None else loss(pred, labels)
        return (loss_value, pred)

    def save_pretrained(self, model_dir, **kwargs):
        """Save model"""
        self.config.save_pretrained(model_dir)
        self.model.save_pretrained(model_dir)
        self.tokenizer.save_pretrained(model_dir)
    
    def load_state_dict(self, model_dir):
        """Load state dict from a directory"""
        return self.model.load_state_dict(self.architecture_class.from_pretrained(model_dir).state_dict())

    def to_pyterrier(self) -> "pt.Transformer":
        return self.transformer_class.from_model(self.model, self.tokenizer, text_field='text')

    @classmethod
    def from_pretrained(cls, model_name_or_path : str, num_labels=2, config=None, **kwargs) -> "Cat":
        """Load model from a directory"""
        config = cls.config_class.from_pretrained(model_name_or_path, num_labels=num_labels) if config is None else config
        model = cls.architecture_class.from_pretrained(model_name_or_path, **kwargs)
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        return cls(model, tokenizer, config)

AutoConfig.register("Cat", CatConfig)
AutoModelForSequenceClassification.register(CatConfig, Cat)


import pyterrier as pt
from transformers import PreTrainedModel, PreTrainedConfig, PreTrainedTokenizer, AutoTokenizer, AutoConfig, AutoModelForSeq2SeqLM, AutoModelForCausalLM
from typing import Union
import torch
import pandas as pd
from more_itertools import chunked
import numpy as np
from .._util import not_tested

DEFAULT_MONO_PROMPT = r'query: {query} document: {text} relevant:'
DEFAULT_DUO_PROMPT = r'query: {query} positive: {text} negative: {text} relevant:'

class Seq2SeqConfig(PreTrainedConfig):
    model_type = 'Seq2Seq'
    @classmethod
    def from_pretrained(cls, model_name_or_path, **kwargs) -> 'Seq2SeqConfig':
        config = super().from_pretrained(model_name_or_path, **kwargs)
        return config

@not_tested
class Seq2SeqTransformer(pt.Transformer):
    architecture_class = AutoModelForSeq2SeqLM
    config_class = Seq2SeqConfig
    def __init__(self, 
                 model : PreTrainedModel, 
                 tokenizer : PreTrainedTokenizer, 
                 config : PreTrainedConfig, 
                 batch_size : int, 
                 text_field : str = 'text', 
                 device : Union[str, torch.device] = None,
                 pos_token : str = 'true',
                 neg_token : str = 'false',
                 prompt : str = None,
                 verbose : bool = False
                 ) -> None:
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.config = config
        self.batch_size = batch_size
        self.text_field = text_field
        self.device = device if device is not None else 'cuda' if torch.cuda.is_available() else 'cpu'
        self.pos_token = self.tokenizer.encode(pos_token)[0]
        self.neg_token = self.tokenizer.encode(neg_token)[0]
        self.prompt = prompt if prompt is not None else DEFAULT_MONO_PROMPT
        self.verbose = verbose
    
    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str, 
                        batch_size : int = 64, 
                        text_field : str = 'text', 
                        device : Union[str, torch.device] = None,
                        prompt : str = None,
                        verbose : bool = False,
                        **kwargs
                        ):
        model = cls.architecture_class.from_pretrained(model_name_or_path, **kwargs).to(device).eval()
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        config = AutoConfig.from_pretrained(model_name_or_path)
        return cls(model, tokenizer, config, batch_size, text_field, device, prompt, verbose=verbose)

    @classmethod 
    def from_model(cls, 
                   model : PreTrainedModel, 
                   tokenizer : PreTrainedTokenizer,
                   batch_size : int = 64, 
                   text_field : str = 'text', 
                   verbose : bool = False
                   ): 
        config = model.config
        return cls(model, tokenizer, config, batch_size, text_field, model.device, verbose=verbose)
    
    def transform(self, inp : pd.DataFrame) -> pd.DataFrame:
        scores = []
        it = inp[['query', self.text_field]].itertuples(index=False)
        if self.verbose:
            it = pt.tqdm(it, total=len(inp), unit='record', desc='Cat scoring')
        with torch.no_grad():
            for chunk in chunked(it, self.batch_size):
                queries, texts = map(list, zip(*chunk))
                prompts = [self.prompt.format(query=q, text=t) for q, t in zip(queries, texts)]
                inps = self.tokenizer(prompts, return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                scores.append(self.model(**inps).logits[:, (self.pos_token, self.neg_token)].softmax(dim=-1)[0].cpu().detach().numpy())
        res = inp.assign(score=np.concatenate(scores))
        pt.model.add_ranks(res)
        res = res.sort_values(['qid', 'rank'])
        return res

class Seq2SeqDuoTransformer(Seq2SeqTransformer):
    architecture_class = AutoModelForSeq2SeqLM
    config_class = Seq2SeqConfig
    def __init__(self,
                    model : PreTrainedModel,
                    tokenizer : PreTrainedTokenizer,
                    config : PreTrainedConfig,
                    batch_size : int,
                    text_field : str = 'text',
                    device : Union[str, torch.device] = None,
                    pos_token : str = 'true',
                    neg_token : str = 'false',
                    prompt : str = None,
                    verbose : bool = False
                    ) -> None:
            raise NotImplementedError("Incomplete, do not use")
            super().__init__(model, tokenizer, config, batch_size, text_field, device, pos_token, neg_token, prompt, verbose)
            self.prompt = prompt if prompt is not None else DEFAULT_DUO_PROMPT

    def transform(self, inp : pd.DataFrame) -> pd.DataFrame:
        scores = []
        it = inp[['query', self.text_field]].itertuples(index=False)
        if self.verbose:
            it = pt.tqdm(it, total=len(inp), unit='record', desc='Cat scoring')
        with torch.no_grad():
            for chunk in chunked(it, self.batch_size):
                queries, texts = map(list, zip(*chunk))
                prompts = [self.prompt.format(query=q, text1=t1, text2=t2) for q, t1, t2 in zip(queries, texts, texts) if t1 != t2]
                inps = self.tokenizer(prompts, return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                scores.append(self.model(**inps).logits[:, (self.pos_token, self.neg_token)].softmax(dim=-1)[0].cpu().detach().numpy())
        res = inp.assign(score=np.concatenate(scores))
        pt.model.add_ranks(res)
        res = res.sort_values(['qid', 'rank'])
        return res
@not_tested
class Seq2Seq(PreTrainedModel):
    """Wrapper for ConditionalGenerationCat Model
    
    Parameters
    ----------
    model : AutoModelForSeq2SeqLM
        the underlying HF model
    config : AutoConfig
        the configuration for the model
    """
    model_type = 'Seq2Seq'
    architecture_class = AutoModelForSeq2SeqLM
    config_class = Seq2SeqConfig
    transformer_class = Seq2SeqTransformer
    def __init__(
        self,
        model: AutoModelForSeq2SeqLM,
        tokenizer: PreTrainedTokenizer,
        config: PreTrainedConfig,
    ):
        super().__init__(config)
        self.model = model
        self.tokenizer = tokenizer

    def prepare_outputs(self, logits):
        raise NotImplementedError

    def forward(self, loss, sequences, labels=None):
        """Compute the loss given (pairs, labels)"""
        sequences = {k: v.to(self.model.device) for k, v in sequences.items()}
        labels = labels.to(self.model.device) if labels is not None else None
        logits = self.model(**sequences).logits
        pred = self.prepare_outputs(logits)
        loss_value = loss(pred) if labels is None else loss(pred, labels)
        return (loss_value, pred)


    def save_pretrained(self, model_dir, **kwargs):
        """Save model"""
        self.config.save_pretrained(model_dir)
        self.model.save_pretrained(model_dir)
        self.tokenizer.save_pretrained(model_dir)
    
    def load_state_dict(self, model_dir):
        """Load state dict from a directory"""
        return self.model.load_state_dict(self.architecture_class.from_pretrained(model_dir).state_dict())
    
    def to_pyterrier(self) -> "Seq2SeqTransformer":
        return self.transformer_class.from_model(self.model, self.tokenizer, text_field='text')

    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str, 
                        config : PreTrainedConfig = None, 
                        **kwargs):
        """Load model from a directory"""
        config = cls.config_class.from_pretrained(model_name_or_path)
        model = cls.architecture_class.from_pretrained(model_name_or_path, **kwargs)
        return cls(model, config)
    
class CausalLMConfig(PreTrainedConfig):
    model_type = 'CausalLM'
    @classmethod
    def from_pretrained(cls, model_name_or_path, **kwargs) -> 'CausalLMConfig':
        config = super().from_pretrained(model_name_or_path, **kwargs)
        return config

class CausalLMTransformer(Seq2SeqTransformer):
    architecture_class = AutoModelForCausalLM
    config_class = CausalLMConfig
    def __init__(self, 
                 model : PreTrainedModel, 
                 tokenizer : PreTrainedTokenizer, 
                 config : PreTrainedConfig, 
                 batch_size : int, 
                 text_field : str = 'text', 
                 device : Union[str, torch.device] = None,
                 prompt : str = None,
                 verbose : bool = False
                 ) -> None:
        raise NotImplementedError("Incomplete, do not use")
        super().__init__(model, tokenizer, config, batch_size, text_field, device, prompt, verbose)
    
    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str, 
                        batch_size : int = 64, 
                        text_field : str = 'text', 
                        config : PreTrainedConfig = None,
                        device : Union[str, torch.device] = None,
                        prompt : str = None,
                        verbose : bool = False,
                        **kwargs
                        ):
        config = cls.config_class.from_pretrained(model_name_or_path) if config is None else config
        model = cls.architecture_class.from_pretrained(model_name_or_path, **kwargs).to(device).eval()
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        return cls(model, tokenizer, config, batch_size, text_field, device, prompt, verbose=verbose)

    @classmethod 
    def from_model(cls, 
                   model : PreTrainedModel, 
                   tokenizer : PreTrainedTokenizer,
                   batch_size : int = 64, 
                   text_field : str = 'text', 
                   verbose : bool = False
                   ): 
        config = model.config
        return cls(model, tokenizer, config, batch_size, text_field, model.device, verbose=verbose)
    
    def transform(self, inp : pd.DataFrame) -> pd.DataFrame:
        scores = []
        it = inp[['query', self.text_field]].itertuples(index=False)
        if self.verbose:
            it = pt.tqdm(it, total=len(inp), unit='record', desc='Cat scoring')
        with torch.no_grad():
            for chunk in chunked(it, self.batch_size):
                queries, texts = map(list, zip(*chunk))
                prompts = [self.prompt.format(query=q, text=t) for q, t in zip(queries, texts)]
                inps = self.tokenizer(prompts, return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                scores.append(self.model(**inps).logits[:, 0].cpu().detach().numpy())
        res = inp.assign(score=np.concatenate(scores))
        pt.model.add_ranks(res)
        res = res.sort_values(['qid', 'rank'])
        return res

class CausalLM(Seq2Seq):
    """Wrapper for CausalLM Model
    
    Parameters
    ----------
    model : AutoModelForCausalLM
        the underlying HF model
    tokenizer : PreTrainedTokenizer
        the tokenizer for the model
    config : AutoConfig
        the configuration for the model
    """
    model_type = 'CausalLM'
    architecture_class = AutoModelForCausalLM
    transformer_class = CausalLMTransformer
    config_class = CausalLMConfig
    def __init__(self, model, tokenizer, config):
        raise NotImplementedError("Incomplete, do not use")
        super().__init__(model, tokenizer, config)

AutoModelForSeq2SeqLM.register("Seq2Seq", Seq2Seq)
AutoModelForCausalLM.register("CausalLM", CausalLM)
AutoConfig.register("Seq2Seq", Seq2SeqConfig)
AutoConfig.register("CausalLM", CausalLMConfig)


from dataclasses import dataclass, field, fields, asdict
from typing import Dict, Any
import json
from enum import Enum
from typing import Optional
import torch
from .. import is_torch_available


@dataclass
class RankerModelArguments:
    model_name_or_path : str = field(
        metadata={"help": "Huggingface model name or path to model"}
    )

    def __str__(self):
        self_as_dict = asdict(self)

        self_as_dict = {k: f"<{k.upper()}>" if k.endswith("_token") else v for k, v in self_as_dict.items()}

        attrs_as_str = [f"{k}={v},\n" for k, v in sorted(self_as_dict.items())]
        return f"{self.__class__.__name__}(\n{''.join(attrs_as_str)})"

    __repr__ = __str__

    def to_dict(self):
        """
        Serializes this instance while replace `Enum` by their values (for JSON serialization support). It obfuscates
        the token values by removing their value.
        """
        # filter out fields that are defined as field(init=False)
        d = {field.name: getattr(self, field.name) for field in fields(self) if field.init}

        for k, v in d.items():
            if isinstance(v, Enum):
                d[k] = v.value
            if isinstance(v, list) and len(v) > 0 and isinstance(v[0], Enum):
                d[k] = [x.value for x in v]
            if k.endswith("_token"):
                d[k] = f"<{k.upper()}>"

        return d

    def to_json_string(self):
        """
        Serializes this instance to a JSON string.
        """
        return json.dumps(self.to_dict(), indent=2)

    def to_sanitized_dict(self) -> Dict[str, Any]:
        """
        Sanitized serialization to use with TensorBoard’s hparams
        """
        d = self.to_dict()

        valid_types = [bool, int, float, str]
        if is_torch_available():
            valid_types.append(torch.Tensor)

        return {k: v if type(v) in valid_types else str(v) for k, v in d.items()}

@dataclass
class RankerDotArguments(RankerModelArguments):
    pooling : Optional[str] = field(
        default='cls',
        metadata={"help": "Pooling strategy"}
    )
    use_pooler : Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to use the pooler MLP"}
    )
    model_tied : Optional[bool] = field(
        default=False,
        metadata={"help": "Whether to tie the weights of the query and document encoder"}
    )
    in_batch_loss : Optional[str] = field(
        default=None,
        metadata={"help": "Loss function to use for in-batch negatives"}
    )

    def __post_init__(self):
        from .loss import LOSS_REGISTRY
        assert self.pooling in ['cls', 'mean', 'none', 'late_interaction'], "Pooling must be one of 'cls', 'mean', 'late_interaction' or 'none'"
        assert self.in_batch_loss is None or self.in_batch_loss in LOSS_REGISTRY.available, f"In-batch loss must be one of {LOSS_REGISTRY.available}"

    

RankerCatArguments = RankerModelArguments


from transformers import PretrainedConfig, PreTrainedModel, PreTrainedTokenizer, AutoModel, AutoTokenizer
from ...modelling.dot import Dot, Pooler, DotConfig
from typing import Union
import torch
import pandas as pd
from more_itertools import chunked
import numpy as np
import pyterrier as pt

class DotTransformer(pt.Transformer):
    cls_architecture = AutoModel
    cls_config = DotConfig
    def __init__(self, 
                 model : PreTrainedModel, 
                 tokenizer : PreTrainedTokenizer, 
                 config : DotConfig, 
                 batch_size : int, 
                 text_field : str = 'text', 
                 device : Union[str, torch.device] = None,
                 verbose : bool = False
                 ) -> None:
        super().__init__()
        self.device = device if device is not None else 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model = model.eval().to(self.device)
        self.tokenizer = tokenizer
        self.config = config
        self.batch_size = batch_size
        self.text_field = text_field
        self.pooling = {
            'mean': lambda x: x.mean(dim=1),
            'cls' : lambda x: x[:, 0],
            'late_interaction': lambda x: x,
            'none': lambda x: x,
        }[config.pooling_type]
        self.verbose = verbose

    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str,
                        batch_size : int = 64,
                        pooling : str = 'cls',
                        config : PretrainedConfig = None,
                        text_field : str = 'text',
                        device : Union[str, torch.device] = None,
                        verbose : bool = False,
                        **kwargs
                        ):
        config = cls.cls_config.from_pretrained(model_name_or_path) if config is None else config
        config.pooling_type = pooling
        pooler = None if not config.use_pooler else Pooler.from_pretrained(model_name_or_path+"/pooler")
        model_d = None if config.model_tied else cls.cls_architecture.from_pretrained(model_name_or_path + "/model_d", **kwargs)
        model_q = cls.cls_architecture.from_pretrained(model_name_or_path, **kwargs)
        model = Dot(model_q, config, model_d, pooler)
        return cls(model, AutoTokenizer.from_pretrained(model_name_or_path), config, batch_size, text_field, device, verbose)
    
    @classmethod 
    def from_model(cls, 
                   model : PreTrainedModel, 
                   tokenizer : PreTrainedTokenizer,
                   batch_size : int = 64, 
                   text_field : str = 'text', 
                   verbose : bool = False
                   ):
        config = model.config
        return cls(model.eval(), tokenizer, config, batch_size, text_field, model.device, verbose)
    
    def encode_queries(self, texts, batch_size=None) -> np.ndarray:
        results = []
        with torch.no_grad():
            for chunk in chunked(texts, batch_size or self.batch_size):
                inps = self.tokenizer(list(chunk), return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                res = self.model._encode_q(**inps)
                results.append(res.cpu().numpy())
        if not results:
            return np.empty(shape=(0, 0))
        return np.concatenate(results, axis=0)

    def encode_docs(self, texts, batch_size=None) -> np.ndarray:
        results = []
        with torch.no_grad():
            for chunk in chunked(texts, batch_size or self.batch_size):
                inps = self.tokenizer(list(chunk), return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                res = self.model._encode_d(**inps)
                results.append(res.cpu().numpy())
        if not results:
            return np.empty(shape=(0, 0))
        return np.concatenate(results, axis=0)
    
    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        columns = set(inp.columns)
        modes = [
            (['qid', 'query', self.text_field], self.scorer),
            (['qid', 'query_vec', self.text_field], self.scorer),
            (['qid', 'query', 'doc_vec'], self.scorer),
            (['qid', 'query_vec', 'doc_vec'], self.scorer),
            (['query'], self.query_model),
            ([self.text_field], self.doc_model),
        ]
        for fields, fn in modes:
            if all(f in columns for f in fields):
                return fn()(inp)
        message = f'Unexpected input with columns: {inp.columns}. Supports:'
        for fields, fn in modes:
            message += f'\n - {fn.__doc__.strip()}: {fields}'
        raise RuntimeError(message)
    
    def query_model(self, verbose=None, batch_size=None) -> pt.Transformer:
        """
        Query encoding
        """
        return BiQuerymodel(self, verbose=verbose, batch_size=batch_size)

    def doc_model(self, verbose=None, batch_size=None) -> pt.Transformer:
        """
        Doc encoding
        """
        return BiDocmodel(self, verbose=verbose, batch_size=batch_size)

    def scorer(self, verbose=None, batch_size=None) -> pt.Transformer:
        """
        Scoring (re-ranking)
        """
        return BiScorer(self, verbose=verbose, batch_size=batch_size)

class BiQuerymodel(pt.Transformer):
    def __init__(self, bi_model: DotTransformer, verbose=None, batch_size=None):
        self.bi_model = bi_model
        self.verbose = verbose if verbose is not None else bi_model.verbose
        self.batch_size = batch_size if batch_size is not None else bi_model.batch_size

    def encode(self, texts, batch_size=None) -> np.array:
        return self.bi_model.encode_queries(texts, batch_size=batch_size or self.batch_size)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        assert all(c in inp.columns for c in ['query'])
        it = inp['query'].values
        it, inv = np.unique(it, return_inverse=True)
        if self.verbose:
            it = pt.tqdm(it, desc='Encoding Queries', unit='query')
        enc = self.encode(it)
        return inp.assign(query_vec=[enc[i] for i in inv])

    def __repr__(self):
        return f'{repr(self.bi_model)}.query_model()'
    
class BiDocmodel(pt.Transformer):
    def __init__(self, bi_model: DotTransformer, verbose=None, batch_size=None, text_field=None):
        self.bi_model = bi_model
        self.verbose = verbose if verbose is not None else bi_model.verbose
        self.batch_size = batch_size if batch_size is not None else bi_model.batch_size
        self.text_field = text_field if text_field is not None else bi_model.text_field

    def encode(self, texts, batch_size=None) -> np.array:
        return self.bi_model.encode_docs(texts, batch_size=batch_size or self.batch_size)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        assert all(c in inp.columns for c in [self.text_field])
        it = inp[self.text_field]
        if self.verbose:
            it = pt.tqdm(it, desc='Encoding Docs', unit='doc')
        return inp.assign(doc_vec=list(self.encode(it)))

    def __repr__(self):
        return f'{repr(self.bi_model)}.doc_model()'

class BiScorer(pt.Transformer):
    def __init__(self, bi_model: DotTransformer, verbose=None, batch_size=None, text_field=None):
        self.bi_model = bi_model
        self.verbose = verbose if verbose is not None else bi_model.verbose
        self.batch_size = batch_size if batch_size is not None else bi_model.batch_size
        self.text_field = text_field if text_field is not None else bi_model.text_field

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        assert 'query_vec' in inp.columns or 'query' in inp.columns
        assert 'doc_vec' in inp.columns or self.text_field in inp.columns
        if 'query_vec' in inp.columns:
            query_vec = inp['query_vec']
        else:
            query_vec = self.bi_model.query_model(batch_size=self.batch_size, verbose=self.verbose)(inp)['query_vec']
        if 'doc_vec' in inp.columns:
            doc_vec = inp['doc_vec']
        else:
            doc_vec = self.bi_model.doc_model(batch_size=self.batch_size, verbose=self.verbose)(inp)['doc_vec']
            if self.bi_model.config.pooling_type == 'late_interaction':
                scores = doc_vec @ query_vec.permute(0, 2, 1)
                scores = scores.max(1).values
                scores = scores.sum(-1)
            else:
                scores = (query_vec * doc_vec).apply(np.sum)
        outp = inp.assign(score=scores)
        return pt.model.add_ranks(outp)

    def __repr__(self):
        return f'{repr(self.bi_model)}.scorer()'
    
__all__ = ['DotTransformer']


import pyterrier as pt
from transformers import PreTrainedModel, PretrainedConfig, PreTrainedTokenizer, AutoModelForSequenceClassification, AutoTokenizer, AutoConfig
from typing import Union
import torch
import pandas as pd
from more_itertools import chunked
import numpy as np
import torch.nn.functional as F

class CatTransformer(pt.Transformer):
    cls_architecture = AutoModelForSequenceClassification
    cls_config = AutoConfig
    def __init__(self, 
                 model : PreTrainedModel, 
                 tokenizer : PreTrainedTokenizer, 
                 config : PretrainedConfig, 
                 batch_size : int, 
                 text_field : str = 'text', 
                 device : Union[str, torch.device] = None,
                 verbose : bool = False
                 ) -> None:
        super().__init__()
        self.model = model.eval()
        self.tokenizer = tokenizer
        self.config = config
        self.batch_size = batch_size
        self.text_field = text_field
        self.device = device if device is not None else 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model = self.model.to(self.device)
        self.verbose = verbose
    
    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str, 
                        batch_size : int = 64, 
                        text_field : str = 'text', 
                        config : PretrainedConfig = None,
                        device : Union[str, torch.device] = None,
                        verbose : bool = False,
                        **kwargs
                        ):
        config = cls.cls_config.from_pretrained(model_name_or_path) if config is None else config
        model = cls.cls_architecture.from_pretrained(model_name_or_path, config=config, **kwargs)
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        return cls(model, tokenizer, config, batch_size, text_field, device, verbose)

    @classmethod 
    def from_model(cls, 
                   model : PreTrainedModel, 
                   tokenizer : PreTrainedTokenizer,
                   batch_size : int = 64, 
                   text_field : str = 'text', 
                   verbose : bool = False
                   ): 
        config = model.config
        return cls(model, tokenizer, config, batch_size, text_field, model.device, verbose)
    
    def transform(self, inp : pd.DataFrame) -> pd.DataFrame:
        scores = []
        it = inp[['query', self.text_field]].itertuples(index=False)
        if self.verbose:
            it = pt.tqdm(it, total=len(inp), unit='record', desc='Cat scoring')
        with torch.no_grad():
            for chunk in chunked(it, self.batch_size):
                queries, texts = map(list, zip(*chunk))
                inps = self.tokenizer(queries, texts, return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.model.device) for k, v in inps.items()}
                scores.append(F.log_softmax(self.model(**inps).logits, dim=-1)[:, 1].cpu().detach().numpy())
        res = inp.assign(score=np.concatenate(scores))
        res = res.sort_values(['qid', 'score'], ascending=[True, False])
        return pt.model.add_ranks(res)

class PairTransformer(pt.Transformer):
    cls_architecture = AutoModelForSequenceClassification
    cls_config = AutoConfig
    def __init__(self, 
                 model : PreTrainedModel, 
                 tokenizer : PreTrainedTokenizer, 
                 config : PretrainedConfig, 
                 batch_size : int, 
                 text_field : str = 'text', 
                 device : Union[str, torch.device] = None,
                 verbose : bool = False
                 ) -> None:
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.config = config
        self.batch_size = batch_size
        self.text_field = text_field
        self.device = device if device is not None else 'cuda' if torch.cuda.is_available() else 'cpu'
        self.verbose = verbose
    
    @classmethod 
    def from_model(cls, 
                   model : PreTrainedModel, 
                   tokenizer : PreTrainedTokenizer,
                   batch_size : int = 64, 
                   text_field : str = 'text', 
                   verbose : bool = False
                   ): 
        config = model.config
        return cls(model, tokenizer, config, batch_size, text_field, model.device, verbose)
    
    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str, 
                        batch_size : int = 64, 
                        text_field : str = 'text', 
                        config : PretrainedConfig = None,
                        device : Union[str, torch.device] = None,
                        verbose : bool = False,
                        **kwargs
                        ):
        config = cls.cls_config.from_pretrained(model_name_or_path) if config is None else config
        model = cls.cls_architecture.from_pretrained(model_name_or_path, config=config, **kwargs).cuda().eval()
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        return cls(model, tokenizer, config, batch_size, text_field, device, verbose)
    
    def transform(self, inp : pd.DataFrame) -> pd.DataFrame:
        scores = []
        it = inp[['query', self.text_field]].itertuples(index=False)
        if self.verbose:
            it = pt.tqdm(it, total=len(inp), unit='record', desc='Duo scoring')
        with torch.no_grad():
            for chunk in chunked(it, self.batch_size):
                queries, texts = map(list, zip(*chunk))
                inps = self.tokenizer(queries, texts, return_tensors='pt', padding=True, truncation=True)
                inps = {k: v.to(self.device) for k, v in inps.items()}
                scores.append(self.model(**inps).logits.cpu().detach().numpy())
        res = inp.assign(score=np.concatenate(scores))
        res = inp.assign(score=np.concatenate(scores))
        res = res.sort_values(['qid', 'score'], ascending=[True, False])
        return pt.model.add_ranks(res)
    
__all__ = ['CatTransformer', 'PairTransformer']


import random
from torch.utils.data import Dataset
import pandas as pd
import torch
from typing import Union
import ir_datasets as irds
import json
from .._util import load_json, initialise_irds_eval
from .corpus import Corpus

class LazyTextLoader:
    def __init__(self, corpus : Union[Corpus, irds.Dataset]) -> None:
        self.docstore = corpus.docs_store()
    
    def __getitem__(self, doc_id):
        if type == list:
            return [doc.text for doc in self.docstore.get_many(doc_id)]
        return self.docstore.get(doc_id).text

class TrainingDataset(Dataset):
    def __init__(self, 
                 training_dataset_file: str, 
                 corpus: Union[Corpus, irds.Dataset],
                 teacher_file: str = None,
                 group_size: int = 2,
                 no_positive: bool = False,
                 lazy_load_text : bool = True,
                 top_k_group : bool = False,
                 precomputed: bool = False,
                 text_field: str = 'text',
                 query_field : str = 'text'
                 ) -> None:
        assert training_dataset_file.endswith('jsonl'), "Training dataset should be a JSONL file and should not be compressed"

        self.training_dataset_file = training_dataset_file
        self.corpus = corpus
        self.teacher_file = teacher_file
        self.group_size = group_size
        self.no_positive = no_positive
        self.lazy_load_text = lazy_load_text
        self.n_neg = self.group_size -1 if not self.no_positive else self.group_size
        self.top_k_group = top_k_group
        self.precomputed = precomputed
        self.text_field = text_field
        self.query_field = query_field

        self._get = self._precomputed_get if self.precomputed else self._standard_get

        self.line_offsets = self._get_line_offsets() 
        super().__init__()
        self.__post_init__()

    def _get_line_offsets(self):
        """Store byte offsets for each line in an uncompressed JSONL file."""
        offsets = []
        with open(self.training_dataset_file, 'r', encoding="utf-8") as f:
            while True:
                offset = f.tell()
                line = f.readline()
                if not line:
                    break
                offsets.append(offset)
        return offsets

    def _get_line_by_index(self, idx):
        """Retrieve a line by index, using offsets for uncompressed files."""
        with open(self.training_dataset_file, 'r', encoding="utf-8") as f:
            f.seek(self.line_offsets[idx])
            return json.loads(f.readline())

    def __post_init__(self):
        assert self.corpus is not None or self.precomputed, "Cannot instantiate a text-based dataset without a lookup"
        
        # Initialize documents and queries from corpus
        if not self.precomputed:
            if not self.lazy_load_text:
                self.docs = pd.DataFrame(self.corpus.docs_iter()).set_index("doc_id")["text"].to_dict()
            else:
                self.docs = LazyTextLoader(self.corpus)

            self.queries = pd.DataFrame(self.corpus.queries_iter()).set_index("query_id")["text"].to_dict()

        # Load teacher data if available
        if self.teacher_file:
            self.teacher = load_json(self.teacher_file)
            self.labels = True
        else:
            self.labels = False

        # Use _get_line_by_index to check multi-negative configuration
        first_entry = self._get_line_by_index(0)
        self.multi_negatives = isinstance(first_entry['doc_id_b'], list)
        total_negs = len(first_entry['doc_id_b']) if self.multi_negatives else 1
        assert self.n_neg <= total_negs, f"Only found {total_negs} negatives, cannot take {self.n_neg} negatives"

    def __len__(self):
        # Length based on line offsets for uncompressed, or generator count for compressed
        return len(self.line_offsets) if self.line_offsets else sum(1 for _ in self._data_generator())
    
    def _teacher(self, query_id, doc_id):
        if doc_id not in self.teacher[query_id]: return 0
        else: return self.teacher[query_id][doc_id]

    def _precomputed_get(self, data):
        query, query_id, doc_id_a, doc_id_b = data[self.query_field], data['query_id'], data['doc_id_a'], data[f"{self.text_field}_a"], data[f"{self.text_field}_b"]

        texts = [doc_id_a] if not self.no_positive else []
        if self.multi_negatives:
            texts.extend(doc_id_b)
        else:
            texts.append(doc_id_b)
        
        return (query, query_id, doc_id_a, doc_id_b, texts)

    def _standard_get(self, data):
        query_id, doc_id_a, doc_id_b = data['query_id'], data['doc_id_a'], data['doc_id_b']
        query = self.queries[str(query_id)]
        texts = [self.docs[str(doc_id_a)]] if not self.no_positive else []

        if self.multi_negatives:
            texts.extend([self.docs[str(doc)] for doc in doc_id_b] if not self.lazy_load_text else self.docs[doc_id_b])
        else:
            texts.append(self.docs[str(doc_id_b)])

        return (query, query_id, doc_id_a, doc_id_b, texts)

    def __getitem__(self, idx):
        # Retrieve the line corresponding to idx
        item = self._get_line_by_index(idx)

        query, query_id, doc_id_a, doc_id_b, texts = self._get(item)

        # Append teacher scores if available
        if self.labels:
            scores = [self._teacher(str(query_id), str(doc_id_a))] if not self.no_positive else []
            if self.multi_negatives:
                scores.extend([self._teacher(str(query_id), str(doc)) for doc in doc_id_b])
            else:
                scores.append(self._teacher(str(query_id), str(doc_id_b)))
            if len(texts) > (self.group_size):
                if self.top_k_group:
                    texts, scores = zip(*sorted(zip(texts, scores), key=lambda x: x[1], reverse=True))
                    return (query, texts[:self.group_size], scores[:self.group_size])
            else:
                texts, scores = zip(*random.sample(list(zip(texts, scores)), self.group_size))
            return (query, texts, scores)
        else:
            if len(texts) > (self.group_size): texts = random.sample(texts, self.group_size)
            return (query, texts)


class EvaluationDataset(Dataset):
    def __init__(self, 
                 evaluation_dataset : pd.DataFrame, 
                 corpus : Union[Corpus, irds.Dataset]
                 ) -> None:
        super().__init__()
        self.evaluation_dataset = evaluation_dataset
        self.corpus = corpus

        self.__post_init__()
    
    def __post_init__(self):
        for column in 'query_id', 'docno', 'score':
            if column not in self.evaluation_dataset.columns: raise ValueError(f"Format not recognised, Column '{column}' not found in dataframe")
        self.docs = pd.DataFrame(self.corpus.docs_iter()).set_index("doc_id")["text"].to_dict()
        self.queries = pd.DataFrame(self.corpus.queries_iter()).set_index("query_id")["text"].to_dict()
        self.qrels = pd.DataFrame(self.corpus.qrels_iter())

        self.evaluation_dataset['text'] = self.evaluation_dataset['docno'].map(self.docs)
        self.evaluation_dataset['query'] = self.evaluation_dataset['query_id'].map(self.queries)

    @classmethod
    def from_irds(cls,
                  ir_dataset : irds.Dataset,
                  ) -> 'EvaluationDataset':
            evaluation_dataset = initialise_irds_eval(ir_dataset)
            return cls(evaluation_dataset, ir_dataset)
    
    def __len__(self):
        return len(self.evaluation_dataset.query_id.unique())


from transformers import PretrainedConfig, PreTrainedModel, PreTrainedTokenizer
from .dot import Dot, Pooler
from .._util import not_tested 

@not_tested
class BGEConfig(PretrainedConfig):
    """Configuration for Dot Model
    
    Parameters
    ----------
    model_name_or_path : str
        the model name or path
    mode : str
        the pooling mode for the model
    encoder_tied : bool
        whether the encoder is tied
    use_pooler : bool
        whether to use the pooler
    pooler_dim_in : int
        the input dimension for the pooler
    pooler_dim_out : int
        the output dimension for the pooler
    pooler_tied : bool
        whether the pooler is tied
    """
    model_architecture = "BGE"
    def __init__(self, 
                 model_name_or_path : str='bert-base-uncased',
                 mode='cls', 
                 inbatch_loss = None,
                 encoder_tied=True,
                 use_pooler=False,
                 pooler_dim_in=768,
                 pooler_dim_out=768,
                 pooler_tied=True,
                 **kwargs):
        self.model_name_or_path = model_name_or_path
        self.mode = mode
        self.inbatch_loss = inbatch_loss
        self.encoder_tied = encoder_tied
        self.use_pooler = use_pooler
        self.pooler_dim_in = pooler_dim_in
        self.pooler_dim_out = pooler_dim_out
        self.pooler_tied = pooler_tied
        super().__init__(**kwargs)
    
    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str='bert-base-uncased',
                        mode='cls', 
                        inbatch_loss = None,
                        encoder_tied=True,
                        use_pooler=False,
                        pooler_dim_in=768,
                        pooler_dim_out=768,
                        pooler_tied=True,
                          ) -> 'BGEConfig':
        config = super().from_pretrained(model_name_or_path)
        config.model_name_or_path = model_name_or_path
        config.mode = mode
        config.inbatch_loss = inbatch_loss
        config.encoder_tied = encoder_tied
        config.use_pooler = use_pooler
        config.pooler_dim_in = pooler_dim_in
        config.pooler_dim_out = pooler_dim_out
        config.pooler_tied = pooler_tied
        return config
    
    
@not_tested
class BGE(Dot):
    """
    Dot Model for Fine-Tuning 

    Parameters
    ----------
    encoder : PreTrainedModel
        the encoder model
    config : DotConfig
        the configuration for the model
    encoder_d : PreTrainedModel
        the document encoder model
    pooler : Pooler
        the pooling layer
    """
    def __init__(
        self,
        encoder : PreTrainedModel,
        tokenizer : PreTrainedTokenizer,
        config : BGEConfig,
        encoder_d : PreTrainedModel = None,
        pooler : Pooler = None,
    ):
        raise NotImplementedError("Incomplete, do not use")
        super().__init__(config)
        self.encoder = encoder
        self.tokenizer = tokenizer
        if encoder_d: self.encoder_d = encoder_d
        else: self.encoder_d = self.encoder if config.encoder_tied else deepcopy(self.encoder)
        self.pooling = {
            'mean': self._mean,
            'cls' : self._cls,
            'none': lambda x: x,
        }[config.mode]

        if config.use_pooler: self.pooler = Pooler(config) if pooler is None else pooler
        else: self.pooler = lambda x, y =True : x

        if config.inbatch_loss is not None:
            if config.inbatch_loss not in loss.__all__:
                raise ValueError(f"Unknown loss: {config.inbatch_loss}")
            self.inbatch_loss_fn = getattr(loss, config.inbatch_loss)()
        else:
            self.inbatch_loss_fn = None


from transformers import TrainingArguments
from transformers.trainer_pt_utils import AcceleratorConfig
from transformers.utils import is_accelerate_available
from dataclasses import field, fields, dataclass
from typing import List, Optional
from enum import Enum
import os
from .. import is_ir_measures_available, seed_everything


def parse_ir_measure(measure : str):
    if not is_ir_measures_available(): return measure
    from ir_measures import parse_measure
    return parse_measure(measure)

def get_loss(loss_fn : str):
    from .loss import LOSS_REGISTRY
    if isinstance(loss_fn, str):
        if loss_fn not in LOSS_REGISTRY.available: raise ValueError(f"Unknown loss: {loss_fn}, choices are {LOSS_REGISTRY.available}")
        return LOSS_REGISTRY.get(loss_fn)
    else:
        return loss_fn

@dataclass
class RankerTrainingArguments(TrainingArguments):
    group_size : Optional[int] = field(
        default=2,
        metadata={"help": "Number of documents per query"}
    )
    eval_metrics : Optional[List[str]] = field(
        default_factory=lambda: [],
        metadata={"help": "Evaluation metrics"}
    )
    wandb_project : Optional[str] = field(
        default=None,
        metadata={"help": "Wandb project name"}
    )
    loss_fn : Optional[str] = field(
        default='lce',
        metadata={"help": "Loss function to use"}
    )

    def __post_init__(self):
        super().__post_init__()
        seed_everything(self.seed)

        if self.wandb_project is not None:
            os.environ['WANDB_PROJECT'] = self.wandb_project
        assert self.group_size > 0, "Group size must be greater than 0"
        
        self.eval_ir_metrics = [parse_ir_measure(metric) for metric in self.eval_metrics] if len(self.eval_metrics) > 0 else None
        self.loss_fn = get_loss(self.loss_fn)


from copy import deepcopy
import os
import torch
from torch import nn
import pyterrier as pt
from transformers import PreTrainedModel, PreTrainedTokenizer, PretrainedConfig, AutoModel, AutoTokenizer, AutoConfig
from ..train.loss import batched_dot_product, cross_dot_product

class DotConfig(PretrainedConfig):
    """Configuration for Dot Model
    
    Parameters
    ----------
    model_name_or_path : str
        the model name or path
    mode : str
        the pooling mode for the model
    model_tied : bool
        whether the model is tied
    use_pooler : bool
        whether to use the pooler
    pooler_dim_in : int
        the input dimension for the pooler
    pooler_dim_out : int
        the output dimension for the pooler
    pooler_tied : bool
        whether the pooler is tied
    """
    model_type = "Dot"
    def __init__(self, 
                 model_name_or_path : str='bert-base-uncased',
                 pooling_type ='cls', 
                 inbatch_loss=None,
                 model_tied=True,
                 use_pooler=False,
                 pooler_dim_in=768,
                 pooler_dim_out=768,
                 pooler_tied=True,
                 **kwargs):
        self.model_name_or_path = model_name_or_path
        self.pooling_type = pooling_type
        self.inbatch_loss = inbatch_loss
        self.model_tied = model_tied
        self.use_pooler = use_pooler
        self.pooler_dim_in = pooler_dim_in
        self.pooler_dim_out = pooler_dim_out
        self.pooler_tied = pooler_tied
        super().__init__(**kwargs)
    
    @classmethod
    def from_pretrained(cls, 
                        model_name_or_path : str='bert-base-uncased',
                        pooling_type ='cls', 
                        inbatch_loss=None,
                        model_tied=True,
                        use_pooler=False,
                        pooler_dim_in=768,
                        pooler_dim_out=768,
                        pooler_tied=True,
                          ) -> 'DotConfig':
        config = super().from_pretrained(model_name_or_path)
        config.model_name_or_path = model_name_or_path
        config.pooling_type = pooling_type
        config.inbatch_loss = inbatch_loss
        config.model_tied = model_tied
        config.use_pooler = use_pooler
        config.pooler_dim_in = pooler_dim_in
        config.pooler_dim_out = pooler_dim_out
        config.pooler_tied = pooler_tied
        return config

class Pooler(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense_q = nn.Linear(config.pooler_dim_in, config.pooler_dim_out)
        self.dense_d = nn.Linear(config.pooler_dim_in, config.pooler_dim_out) if not config.pooler_tied else self.dense_q
    
    @classmethod
    def from_pretrained(cls, model_name_or_path : str='bert-base-uncased') -> 'Pooler':
        config = DotConfig.from_pretrained(model_name_or_path)
        model = cls(config)
        return model
    
    def forward(self, hidden_states, d=False):
        return self.dense_d(hidden_states) if d else self.dense_q(hidden_states)

class Dot(PreTrainedModel):
    """
    Dot Model for Fine-Tuning 

    Parameters
    ----------
    model : PreTrainedModel
        the model model
    config : DotConfig
        the configuration for the model
    model_d : PreTrainedModel
        the document model model
    pooler : Pooler
        the pooling layer
    """
    model_type = "Dot"
    architecture_class = AutoModel
    config_class = DotConfig
    transformer_class = None
    def __init__(
        self,
        model : PreTrainedModel,
        tokenizer : PreTrainedTokenizer,
        config : DotConfig,
        model_d : PreTrainedModel = None,
        pooler : Pooler = None,
    ):
        super().__init__(config)
        self.model = model
        self.tokenizer = tokenizer
        if model_d: self.model_d = model_d
        else: self.model_d = self.model if config.model_tied else deepcopy(self.model)
        self.pooling = {
            'mean': lambda x: x.mean(dim=1),
            'cls' : lambda x: x[:, 0],
            'late_interaction': lambda x: x,
            'none': lambda x: x,
        }[config.pooling_type]
        self.pooling_type = config.pooling_type

        if config.use_pooler: self.pooler = Pooler(config) if pooler is None else pooler
        else: self.pooler = lambda x, y =True : x

        self.inbatch_loss_fn = config.inbatch_loss

        from .pyterrier.dot import DotTransformer
        self.transformer_class = DotTransformer


    def prepare_outputs(self, query_reps, docs_batch_reps, labels=None):
        batch_size = query_reps.size(0)

        if self.pooling_type == 'late_interaction':
            pred = emb_q @ emb_d.permute(0, 2, 1)
            pred = pred.max(1).values
            pred = pred.sum(-1)
        else:
            emb_q = query_reps.reshape(batch_size, 1, -1)
            emb_d = docs_batch_reps.reshape(batch_size, self.config.group_size, -1)
            pred = batched_dot_product(emb_q, emb_d)

        if self.config.inbatch_loss is not None:
            if self.pooling_type == 'late_interaction':
                inbatch_d = emb_d[:, 0]
                inbatch_pred = emb_q @ inbatch_d.permute(0, 2, 1)
                inbatch_pred = inbatch_pred.max(1).values
                inbatch_pred = inbatch_pred.sum(-1)
            else:
                inbatch_d = emb_d[:, 0]
                inbatch_pred = cross_dot_product(emb_q.view(batch_size, -1), inbatch_d)
        else:
            inbatch_pred = None

        if labels is not None: labels = labels.reshape(batch_size, self.config.group_size)

        return pred, labels, inbatch_pred
    
    def _cls(self, x : torch.Tensor) -> torch.Tensor:
        return self.pooler(x[:, 0])
    
    def _mean(self, x : torch.Tensor) -> torch.Tensor:
        return self.pooler(x.mean(dim=1))
    
    def _encode_d(self, **text):
        return self.pooling(self.model_d(**text).last_hidden_state)
    
    def _encode_q(self, **text):
        return self.pooling(self.model(**text).last_hidden_state)

    def forward(self, 
                loss = None, 
                queries = None, 
                docs_batch = None, 
                labels=None):
        """Compute the loss given (queries, docs, labels)"""
        queries = {k: v.to(self.model.device) for k, v in queries.items()} if queries is not None else None
        docs_batch = {k: v.to(self.model_d.device) for k, v in docs_batch.items()} if docs_batch is not None else None
        labels = labels.to(self.model_d.device) if labels is not None else None

        query_reps = self._encode_q(**queries) if queries is not None else None
        docs_batch_reps = self._encode_d(**docs_batch) if docs_batch is not None else None

        pred, labels, inbatch_pred = self.prepare_outputs(query_reps, docs_batch_reps, labels)
        inbatch_loss = self.inbatch_loss_fn(inbatch_pred, torch.eye(inbatch_pred.shape[0]).to(inbatch_pred.device)) if inbatch_pred is not None else 0.
        
        loss_value = loss(pred, labels) if labels is not None else loss(pred)
        loss_value += inbatch_loss
        return (loss_value, pred) 

    def save_pretrained(self, model_dir, **kwargs):
        """Save both query and document model"""
        self.config.save_pretrained(model_dir)
        self.model.save_pretrained(model_dir)
        if not self.config.model_tied: self.model_d.save_pretrained(model_dir + "/model_d")
        if self.config.use_pooler: self.pooler.save_pretrained(model_dir + "/pooler")
        self.tokenizer.save_pretrained(model_dir)


    def load_state_dict(self, model_dir):
        """Load state dict from a directory"""
        self.config = DotConfig.from_pretrained(model_dir)
        self.model.load_state_dict(self.architecture_class.from_pretrained(model_dir).state_dict())
        if not self.config.model_tied: self.model_d.load_state_dict(self.architecture_class.from_pretrained(model_dir + "/model_d").state_dict())
        if self.config.use_pooler: self.pooler.load_state_dict(self.architecture_class.from_pretrained(model_dir + "/pooler").state_dict())

    @classmethod
    def from_pretrained(cls, model_name_or_path, config = None, **kwargs) -> "Dot":
        """Load model"""
        if os.path.isdir(model_name_or_path):
            config = cls.config_class.from_pretrained(model_name_or_path) if config is None else config
            model = cls.architecture_class.from_pretrained(model_name_or_path, **kwargs)
            tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
            model_d = None if config.model_tied else cls.architecture_class.from_pretrained(model_name_or_path + "/model_d", **kwargs) 
            pooler = None if not config.use_pooler else Pooler.from_pretrained(model_name_or_path + "/pooler")

            return cls(model, tokenizer, config, model_d, pooler)
        config = cls.config_class(model_name_or_path, **kwargs) if config is None else config
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = cls.architecture_class.from_pretrained(model_name_or_path)
        return cls(model, tokenizer, config)
    
    def to_pyterrier(self) -> "DotTransformer":
        return self.transformer_class.from_model(self, self.tokenizer, text_field='text')

AutoConfig.register("Dot", DotConfig)
AutoModel.register(DotConfig, Dot)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rankers",
    version="0.0.4",
    author="Andrew Parry",
    description="A framework for training and evaluating neural IR models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
)



from dataclasses import dataclass, field, fields
from typing import Dict, Any, Optional
import json
from enum import Enum
import torch
from .. import is_ir_datasets_available, is_torch_available, is_pyterrier_available
import pandas as pd
import logging

logging.basicConfig(level=logging.INFO)

@dataclass
class RankerDataArguments:
    training_dataset_file : str = field(
        metadata={"help": "Path to the training dataset in a JSONL format (Must be uncompressed)"}
    )
    teacher_file : Optional[str] = field(
        default=None,
        metadata={"help": "Path to the teacher scores in a JSON format"}
    )
    validation_dataset_file : Optional[str] = field(
        default=None,
        metadata={"help": "Path to the validation dataset in a TREC format"}
    )
    test_dataset_file : Optional[str] = field(
        default=None,
        metadata={"help": "Path to the test dataset in a TREC format"}
    )
    ir_dataset : Optional[str] = field(
        default=None,
        metadata={"help": "IR Dataset for text lookup"}
    )
    no_positive : Optional[bool] = field(
        default=False,
        metadata={"help": "Dont use positive samples locatd in 'doc_id_a' column instead use solely 'doc_id_b'"}
    )
    top_k_group : Optional[bool] = field(
        default=False,
        metadata={"help": "Sort by score when doing list-wise ranking and take top-k"}
    )
    eval_ir_dataset : Optional[str] = field(
        default=None,
        metadata={"help": "IR Dataset for evaluation"}
    )

    def __post_init__(self):
        if self.ir_dataset is not None:
            assert is_ir_datasets_available(), "Please install ir_datasets to use the ir_dataset argument"
            try:
                import ir_datasets
                self.ir_dataset = ir_datasets.load(self.ir_dataset)
            except Exception as e:
                raise ValueError(f"Unable to load ir_dataset: {e}")
        if self.eval_ir_dataset is not None:
            assert is_ir_datasets_available(), "Please install ir_datasets to use the eval_ir_dataset argument"
            try:
                import ir_datasets
                self.eval_ir_dataset = ir_datasets.load(self.eval_ir_dataset)
            except Exception as e:
                raise ValueError(f"Unable to load eval_ir_dataset: {e}")
        assert self.training_dataset_file.endswith('jsonl') or self.training_dataset_file.endswith('jsonl.gz'), "Training dataset should be a JSONL file"
        if self.teacher_file:
            assert self.teacher_file.endswith('json') or self.teacher_file.endswith('json.gz'), "Teacher file should be a JSON file"
        if self.validation_dataset_file:
            assert self.validation_dataset_file.endswith(".gz") or self.validation_dataset_file.endswith(".tsv") or self.validation_dataset_file.endswith(".rez"), "Validation dataset should be a TREC formatted run file"
            if is_pyterrier_available():
                import pyterrier as pt
                self.validation_data = pt.io.read_results(self.validation_dataset_file)
            else:
                logging.warning("Pyterrier not available, validation dataset will be loaded as a DataFrame")
                self.validation_data= pd.read_csv(self.validation_dataset_file, sep="\t")
        if self.test_dataset_file:
            assert self.test_dataset_file.endswith(".gz") or self.test_dataset_file.endswith(".tsv") or self.test_dataset_file.endswith(".rez"), "Test dataset should be a TREC formatted run file"
            if is_pyterrier_available():
                import pyterrier as pt
                self.test_data = pt.io.read_results(self.test_dataset_file)
            else:
                logging.warning("Pyterrier not available, test dataset will be loaded as a DataFrame")
                self.test_data = pd.read_csv(self.test_dataset_file, sep="\t")

        
    def to_dict(self):
        """
        Serializes this instance while replace `Enum` by their values (for JSON serialization support). It obfuscates
        the token values by removing their value.
        """
        # filter out fields that are defined as field(init=False)
        d = {field.name: getattr(self, field.name) for field in fields(self) if field.init}

        for k, v in d.items():
            if isinstance(v, Enum):
                d[k] = v.value
            if isinstance(v, list) and len(v) > 0 and isinstance(v[0], Enum):
                d[k] = [x.value for x in v]
            if k.endswith("_token"):
                d[k] = f"<{k.upper()}>"

        return d

    def to_json_string(self):
        """
        Serializes this instance to a JSON string.
        """
        return json.dumps(self.to_dict(), indent=2)

    def to_sanitized_dict(self) -> Dict[str, Any]:
        """
        Sanitized serialization to use with TensorBoard’s hparams
        """
        d = self.to_dict()
        
        valid_types = [bool, int, float, str]
        if is_torch_available():
            valid_types.append(torch.Tensor)

        return {k: v if type(v) in valid_types else str(v) for k, v in d.items()}


import torch.nn as nn
import torch
from torch import Tensor
import functools

class SingletonMeta(type):
    """
    Metaclass to implement the Singleton design pattern.
    Ensures only one instance of a class can be created.
    """
    _instances = {}
    
    def __call__(cls, *args, **kwargs):
        """
        Controlled object creation to ensure only one instance exists.
        """
        if cls not in cls._instances:
            # If no instance exists, create one
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

class LossFunctionRegistry(metaclass=SingletonMeta):
    """
    A singleton registry for managing and retrieving loss functions by name.
    Supports both built-in PyTorch losses and custom loss functions.
    """
    
    def __init__(self):
        # Check if the registry has already been initialized
        if not hasattr(self, '_registry'):
            # Dictionary to store registered loss functions
            self._registry = {}
            
            # Automatically register built-in PyTorch loss functions
            self._register_builtin_losses()
    
    def _register_builtin_losses(self):
        """
        Automatically register common PyTorch loss functions.
        """
        builtin_losses = {
            # Basic losses
            'mse': nn.MSELoss,
            'l1': nn.L1Loss,
            'cross_entropy': nn.CrossEntropyLoss,
            'nll': nn.NLLLoss,
            'binary_cross_entropy': nn.BCELoss,
            'binary_cross_entropy_with_logits': nn.BCEWithLogitsLoss,
            
            # Reduction variant losses
            'mse_sum': functools.partial(nn.MSELoss, reduction='sum'),
            'mse_none': functools.partial(nn.MSELoss, reduction='none'),
        }
        
        for name, loss_fn in builtin_losses.items():
            self.register(name, loss_fn)
    
    def register(self, name, loss_fn):
        """
        Register a loss function with a given name.
        
        Args:
            name (str): Name to register the loss function under
            loss_fn (callable): Loss function to register
        """
        if not callable(loss_fn):
            raise TypeError(f"Loss function {name} must be callable")
        
        if name in self._registry:
            print(f"Warning: Overwriting existing loss function '{name}'")
        
        self._registry[name] = loss_fn
    
    def get(self, name, **kwargs):
        """
        Retrieve a loss function by name.
        
        Args:
            name (str): Name of the loss function
            **kwargs: Additional arguments to pass to the loss function constructor
        
        Returns:
            callable: Instantiated loss function
        
        Raises:
            KeyError: If the loss function is not found in the registry
        """
        if name not in self._registry:
            available_losses = ", ".join(sorted(self._registry.keys()))
            raise KeyError(f"Loss function '{name}' not found. Available losses: {available_losses}")
        
        return self._registry[name](**kwargs)
    
    @property
    def available(self):
        """
        List all available loss functions.
        
        Returns:
            list: Names of registered loss functions
        """
        return list(self._registry.keys())

# Global singleton instance
LOSS_REGISTRY = LossFunctionRegistry()

def register_loss(name):
    """
    Decorator to register a custom loss function.
    
    Args:
        name (str): Name to register the loss function under
    
    Returns:
        decorator function
    """
    def decorator(loss_fn):
        LOSS_REGISTRY.register(name, loss_fn)
        return loss_fn
    return decorator

class BaseLoss(nn.Module):
    """
    Base class for Losses

    Parameters
    ----------
    reduction: str
        the reduction type
    """
    def __init__(self, reduction : str = 'mean') -> None:
        super(BaseLoss, self).__init__()
        self.reduction = reduction
    
    def _reduce(self, a : torch.Tensor):
        return reduce(a, self.reduction)

    def forward(self, *args, **kwargs):
        raise NotImplementedError

def reduce(a : torch.Tensor, reduction : str):
    """
    Reducing a tensor along a given dimension.
    Parameters
    ----------
    a: torch.Tensor
        the input tensor
    reduction: str
        the reduction type
    Returns
    -------
    torch.Tensor
        the reduced tensor
    """
    if reduction == 'none':
        return a
    if reduction == 'mean':
        return a.mean()
    if reduction == 'sum':
        return a.sum()
    if reduction == 'batchmean':
        return a.mean(dim=0).sum()
    raise ValueError(f"Unknown reduction type: {reduction}")

def normalize(a: Tensor, dim: int = -1):
    """
    Normalizing a tensor along a given dimension.
    Parameters
    ----------
    a: torch.Tensor
        the input tensor
    dim: int
        the dimension to normalize along
    Returns
    -------
    torch.Tensor
        the normalized tensor
    """
    min_values = a.min(dim=dim, keepdim=True)[0]
    max_values = a.max(dim=dim, keepdim=True)[0]
    return (a - min_values) / (max_values - min_values + 1e-10)

def residual(a : Tensor):
    """
    Calculating the residual between a positive sample and multiple negatives.
    Parameters
    ----------
    a: torch.Tensor
        the input tensor
    Returns
    -------
    torch.Tensor
        the residuals
    """
    if a.size(1) == 1: return a
    if len(a.size()) == 3:
        assert a.size(2) == 1, "Expected scalar values for residuals."
        a = a.squeeze(2)

    positive = a[:, 0]
    negative = a[:, 1]

    return positive - negative

def dot_product(a: Tensor, b: Tensor):
    """
    Calculating row-wise dot product between two tensors a and b.
    a and b must have the same dimensionality.
    Parameters
    ----------
    a: torch.Tensor
        size: batch_size x vector_dim
    b: torch.Tensor
        size: batch_size x vector_dim
    Returns
    -------
    torch.Tensor: size of (batch_size x 1)
        dot product for each pair of vectors
    """
    return (a * b).sum(dim=-1)


def cross_dot_product(a: Tensor, b: Tensor):
    """
    Calculating the cross doc product between each row in a with every row in b. a and b must have the same number of columns, but can have varied nuber of rows.
    Parameters
    ----------
    a: torch.Tensor
        size: (batch_size_1,  vector_dim)
    b: torch.Tensor
        size: (batch_size_2, vector_dim)
    Returns
    -------
    torch.Tensor: of size (batch_size_1, batch_size_2) where the value at (i,j) is dot product of a[i] and b[j].
    """
    return torch.mm(a, b.transpose(0, 1))

def batched_dot_product(a: Tensor, b: Tensor):
    """
    Calculating the dot product between two tensors a and b.

    Parameters
    ----------
    a: torch.Tensor
        size: batch_size x vector_dim
    b: torch.Tensor
        size: batch_size x group_size x vector_dim
    Returns
    -------
    torch.Tensor: size of (batch_size x group_size)
        dot product for each group of vectors
    """
    if len(b.shape) == 2:
        return torch.matmul(a, b.transpose(0, 1))

    # Ensure `a` is of shape (batch_size, 1, vector_dim)
    if len(a.shape) == 2:
        a = a.unsqueeze(1)
    
    # Compute batched dot product, result shape: (batch_size, 1, group_size)
    return torch.bmm(b, a.transpose(1, 2)).squeeze()

def maxsim(a : Tensor, b : Tensor, a_mask : Tensor, temperature : float = 1.0):
    scores = torch.einsum('qin,pjn->qipj', a, b)
    scores, _ = scores.max(-1)
    scores = scores.sum(1) / a_mask[:, 1:].sum(-1, keepdim=True)
    scores = scores / temperature
    return scores

def num_non_zero(a: Tensor):
    """
    Calculating the average number of non-zero columns in each row.
    Parameters
    ----------
    a: torch.Tensor
        the input tensor
    """
    return (a > 0).float().sum(dim=1).mean()
    
from . import listwise as listwise
from . import pointwise as pointwise
from . import pairwise as pairwise

from .listwise import *
from .pointwise import *
from .pairwise import *

__all__ = [*listwise.__all__, *pointwise.__all__, *pairwise.__all__]